package Org::Ruka::IMAPClient;

use strict;
use warnings;
use Coro::Semaphore;
use Data::Dumper;
use base qw 'Net::IMAP::Client';

our %UID_COMMANDS = map { $_ => 1 } qw(COPY FETCH STORE SEARCH SORT THREAD);


sub new {
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->{'readlock'} = Coro::Semaphore->new;
    $self->{'writelock'} = Coro::Semaphore->new;
    $self->{'UID_COMMANDS'} = \%UID_COMMANDS;
    return $self;
}


# SUPER turns on socket options, which doesn't work 
# with weird socket stuff that Coro is doing
sub _get_socket {
    my $self = shift;
#    warn "in get_socket, connected:" . $self->{socket}->connected;
    if (!$self->{socket}) {
        warn "super socket";
        return $self->SUPER::_get_socket(@_);
    }
    return $self->{socket};
}

#dirty hack
sub _reconnect_if_needed {
    return 0;
}

#since Net::IMAP::Client doesn't know it's being parallelized,
#we need to handle locking
#each write needs to have its reads in the right order, but
#multiple writes can be outstanding, if continuation reads are not needed


sub _tell_imap2 {
    my ($self, @cmd) = @_;
    
    my @coros = ();
    foreach (@cmd) {
        my ($cmd, $args) = split / /, $_, 2;
        my $coro = new Coro sub {
            my @data = $self->_tell_imap($cmd, $args);
            return @data;
        };
        push @coros, $coro;
        $coro->ready;
        
    }
    my @ret = ();
    foreach (@coros) {
        push @ret, [ $_->join ];
    }
    return \@ret;
}

sub _tell_imap {
    #my ($ok, $lineparts) = $self->SUPER::_tell_imap(@_);
    # begin copy of _tell_imap from SUPER, to fix one bug
    # did not get a response from maintainer

    my ($self, $cmd, $args, $do_notf) = @_;
    $cmd = uc $cmd;

    my ($lineparts, $ok, $res, $code);

  RETRY1: {
        $self->_send_cmd($cmd, $args);
        redo RETRY1 if $self->_reconnect_if_needed;
    }

    $lineparts = [];
    while ($res = $self->_socket_getline) {
        # print STDERR ">>>>$res<<<<<\n";
        if ($res =~ /^\*/) {
            push @$lineparts, []; # this is a new line interesting in itself
        }
        if ($res =~ /(.*)\{(\d+)\}\r\n/) {
            my ($line, $len) = ($1, $2 + 0);
            push @{$lineparts->[-1]},
              $line,
                $self->_read_literal($len);
        } else {
            ($ok, $code) = $self->_cmd_ok($res);
            if (defined($ok)) {
                last;
            } else {
                push @{$lineparts->[-1]}, $res;
            }
        }
    }
    unless (defined $res) {
        goto RETRY1 if $self->_reconnect_if_needed(1);
    }

    if ($do_notf) {
        no warnings 'uninitialized';
        for (my $i = scalar(@$lineparts); --$i >= 0;) {
            my $line = $lineparts->[$i];

            # 1. notifications don't contain literals
            next if scalar(@$line) != 1;  #this line is changed by russor

            my $text = $line->[0];

            # 2. FETCH notifications only contain FLAGS.  We make a
            #    promise never to FETCH flags alone intentionally.

            # 3. Other notifications will have a first token different
            #    from the running command

            if ( $text =~ /^\*\s+\d+\s+FETCH\s*\(\s*FLAGS\s*\([^\)]*?\)\)/
                   || $text !~ /^\*\s+(?:\d+\s+)?$cmd/ ) {
                my $tokens = Net::IMAP::Client::_parse_tokens($line);
                if ($self->_handle_notification($tokens, 1)) {
                    splice @$lineparts, $i, 1;
                }
                next;
            }

            #last;
        }
    }
    
    
    # end copy
    $self->{'readlock'}->up;   
    return wantarray ? ($ok, $lineparts, $code) : $ok ? $lineparts : undef;
}

sub _send_cmd {
    my ($self, $cmd, $args) = @_;

    local $\;
    use bytes;
    if (defined($self->next_cmd_uid_mode)) {
        if ($self->next_cmd_uid_mode && exists($self->{'UID_COMMANDS'}{$cmd})) {
            $cmd = "UID $cmd";
        }
        $self->next_cmd_uid_mode(undef);
    } elsif ($self->uid_mode && exists($self->{'UID_COMMANDS'}{$cmd})) {
        $cmd = "UID $cmd";
    }
    
    my @literals = ();
    if (ref $args eq 'ARRAY') {
        # may contain literals
        foreach (@$args) {
            if (ref $_ eq 'SCALAR') {
                push @literals, $_;
                $_ = '{' . length($$_) . "}\r\n";
            }
        }
        $args = join('', @$args);
    }

    $self->{'writelock'}->down;
    my $id   = $self->_get_next_id;
    my $socket = $self->_get_socket;
    if (@literals == 0) {
        $cmd = "NIC$id $cmd" . ($args ? " $args" : '') . "\r\n";
        $self->_socket_write($cmd);
        $self->{'readlock'}->down;
    } else {
        $cmd = "NIC$id $cmd ";
        $self->_socket_write($cmd);
        $self->{'readlock'}->down;
        my @split = split(/\r\n/, $args);

        my $ea = Net::IMAP::Client::each_array(@split, @literals);
        while (my ($tmp, $lit) = $ea->()) {
            $self->_socket_write($tmp . "\r\n");
            my $line = $self->_socket_getline;
            # print STDERR "$line - $tmp\n";
            if ($line =~ /^\+/) {
                $self->_socket_write($$lit);
            } else {
                $self->{_error} = "Expected continuation, got: $line";
                # XXX: it's really bad if we get here, what to do?
                return undef;
            }
        }
        $self->_socket_write("\r\n"); # end of command!
    }
    $socket->flush;
    $self->{'writelock'}->up;
    return "NIC$id";

}

sub _cmd_ok {
    my ($self, $res, $id) = @_;
    if (defined($id)) {
        die "Did not expect to get \$id: $id";
    }
    # taken from SUPER::_cmd_ok2, but changed to also retrieve the 'response code'

    if ($res =~ /^NIC\d+\s+OK (\[[^\]]*\])?/i) {
        return (1, $2);
    } elsif ($res =~ /^NIC\d+\s+(?:NO|BAD)(?:\s+(.+))?/i) {
        my $error = $1 || 'unknown error';
        $self->{_error} = $error;
        return (0, $error);
    }
    return ();
}

sub _cmd_ok2 {
    die "should not be in _cmd_ok2!";
}


sub fetch_with_seq {
    my ($self, $msg, $keys) = @_;
    my $wants_many = undef;
    if (ref $msg eq 'ARRAY') {
        $msg = join(',', @$msg);
        $wants_many = 1;
    }
    if (ref $keys eq 'ARRAY') {
        $keys = join(' ', @$keys);
    }
    $self->next_cmd_uid_mode(0);
    my ($ok, $lp) = $self->_tell_imap(FETCH => qq[$msg ($keys)], 1);
    if ($ok) {
        my @ret;
        foreach (@$lp) {
            my $tokens = Net::IMAP::Client::_parse_tokens($_);
            push @{$tokens->[3]}, 'seq_id', $tokens->[1];
            push @ret, { @{$tokens->[3]} };
        }
        return $wants_many || @ret > 1 ? \@ret : $ret[0];
    }
}

sub next_cmd_uid_mode {
    my ($self, $val) = @_;
    if (@_ == 2) {
        return $self->{next_cmd_uid_mode} = $val;
    } else {
        return $self->{next_cmd_uid_mode};
    }
}

# return more info if people want it... 
# (eventually) 
sub copy {
    my ($self, $msg, $folder) = @_;
    my $quoted = $folder;
    Net::IMAP::Client::_string_quote($quoted);
    if (ref $msg eq 'ARRAY') {
        $msg = join(',', @$msg);
    }
    my ($ok, $lineparts, $code) = $self->_tell_imap(COPY => "$msg $quoted", 1);
    return wantarray ? ($ok, $lineparts, $code) : $ok;
}

sub uid_expunge {
    my ($self, @uids) = @_;
    my $ok = $self->_tell_imap('UID EXPUNGE', join(',', @uids), 1);
    return $ok;
}

1;
