package Org::Ruka::IMAPPool;
use strict;
use warnings;
use Coro::Semaphore;
use Org::Ruka::IMAPClient;
use Org::Ruka::HandleSocket;
use Data::Dumper;

use base 'Class::Accessor';

__PACKAGE__->mk_accessors(qw(connection user password idlepool selectedpool connected poollimit folderlru));

sub new {
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    if (!defined($self->poollimit)) {
        $self->poollimit(3);
    }
    $self->folderlru([]);
    $self->connected(0);
    $self->idlepool([]);
    $self->selectedpool({});
    $self->{'add_idle_lock'} = Coro::Semaphore->new;
    $self->{'select_lock'} = Coro::Semaphore->new;
    if (!$self->add_idle()) {
        return undef;
    }
    return $self;
}

sub add_idle {
    my $self = shift;
    
    my $handle = Org::Ruka::HandleSocket->new(@{$self->connection})
        or die "Couldn't open handle";

    my $client = Org::Ruka::IMAPClient->new('socket' => $handle);

    if ($client->login($self->user, $self->password)) {
        # call capability so it's cached
        my $foo = $client->capability();
        push @{$self->idlepool}, $client;
        $self->connected($self->connected + 1);
        warn "Connection: " . $self->connected;
        return 1;
    }
    warn "Couldn't login: " . $client->last_error() . "\n";
    return 0;
}

sub separator {
    my $self = shift;
    $self->check_idle() or die "no idle servers";
    return $self->idlepool->[0]->separator();
}

sub folders_more {
    my $self = shift;
    $self->check_idle() or die "no idle servers";
    return $self->idlepool->[0]->folders_more();
}

sub status {
    my $self = shift;
    $self->check_idle() or die "no idle servers";
    return $self->idlepool->[0]->status(@_);
}

sub check_idle {
    my $self = shift;
    $self->{'add_idle_lock'}->down;
    if (!@{$self->idlepool}) {
        if ($self->poollimit < 1 or $self->connected < $self->poollimit) {
            $self->add_idle();
        } else {
            my $folder = shift @{$self->folderlru};
            if (defined($folder) and defined($self->selectedpool->{$folder})) {
                warn "Reusing connection for $folder";
                my $folderhash = $self->selectedpool->{$folder};
                push @{$self->idlepool}, $folderhash->{'client'};
                delete $self->selectedpool->{$folder};
            } else {
                if (defined($folder)) {
                    warn "Couldn't reuse $folder";
                }
                warn "Couldn't find a connection to reuse but over pool limit";
                $self->add_idle();            
            }
        }
    }
    $self->{'add_idle_lock'}->up;
    return scalar(@{$self->idlepool});
}

sub get_selected_client {
    my $self = shift;
    my ($folder) = @_;
    my $hashref = $self->selectedpool->{$folder};
    my $client;
    if ($hashref->{'client'}) {
        $client = $hashref->{'client'};
    } else {
        my $guard = $self->{'select_lock'}->guard;
        $self->check_idle() or die "no idle servers";
        $client = pop @{$self->idlepool};
        if ($client->select($folder)) {
            $self->selectedpool->{$folder} = { 'client' => $client };
        } else {
            push @{$self->idlepool}, $client;
            return 0;
        }
    }
    for (my $i = 0; $i < @{$self->folderlru}; ++$i) {
        if ($self->folderlru->[$i] eq $folder) {
            splice (@{$self->folderlru}, $i, 0);
            last;
        }
    }
    push @{$self->folderlru}, $folder;
    return $client;
}        

sub unselect {
    my $self = shift;
    my ($folder) = @_;
    for (my $i = 0; $i < @{$self->folderlru}; ++$i) {
        if ($self->folderlru->[$i] eq $folder) {
            splice (@{$self->folderlru}, $i, 0);
            last;
        }
    }
    push @{$self->folderlru}, $folder;
}


1;
