package Org::Ruka::ProxyMailbox;
use strict;
use warnings;

use Org::Ruka::ProxyMessage;
use Data::Dumper;
use Coro;
use Coro::Semaphore;
use Storable qw(freeze nstore retrieve);
use Digest::SHA1 qw(sha1_hex);

use base qw 'Net::IMAP::Server::Mailbox';
Org::Ruka::ProxyMailbox->mk_accessors(qw(imappool refcount selectcount messagecount recentcount unseencount listflags lock needsheaders needsbody lastnoop));

# TODOS:
# notifications
# alterations from client
our $TIME_BETWEEN_NOOPS = 2;
our $PREFETCH_HEADER_COUNT = 50;
our $PREFETCH_BODY_SIZE = 25000;
our $PREFETCH_BODY_SIZE_GUESS = 5000; #i think an average message is about 5 k


sub init {
    my $self = shift;

    # blank starting data...
    $self->children([]);
    $self->messages([]);
    $self->uids({});
    $self->is_selectable(1);
    $self->subscribed(1);
    $self->lock(new Coro::Semaphore);
    $self->needsheaders([]);
    $self->needsbody([]);
    $self->lastnoop(0);
    
    if (ref $self->listflags eq 'ARRAY') {
        foreach (@{$self->listflags}) {
            if ($_ eq '\\Noselect') {
                warn $self->full_path . " is unselectable";
                $self->is_selectable(0);
                $self->subscribed(0);
            }
        }
    }
}

sub add_message {
    my ($self, $message) = @_;
    $self->SUPER::add_message($message);
    push @{$self->needsheaders}, $message->uid;
    push @{$self->needsbody}, $message->uid;
}

sub uids_to_prefetch_headers {
    my ($self, $uid, $count) = @_;
    my @ret;
    if ($uid) { 
        @ret= ($uid);
        --$count;
    }
    my $putback = 0;
    while ($count > 0 and $uid = pop @{$self->needsheaders}) {
        if (!$self->uids->{$uid}->headerloaded) {
            if (!$self->uids->{$uid}->lock->try) {
                # put back into list, but first... 
                unshift @{$self->needsheaders}, 0 unless $putback;
                $putback = 1;
                unshift @{$self->needsheaders}, $uid;
                next;
            }
            push @ret, $uid;
            --$count;
        }
    }
    return sort { $a <=> $b } (@ret);
}


sub uids_to_prefetch_body {
    my ($self, $uid, $size) = @_;
    my @ret;
    if ($uid) {
        @ret = ($uid);
        if (defined($self->uids->{$uid}->_size)) {
            $size -= $self->uids->{$uid}->_size;
        } else {
            $size -= $PREFETCH_BODY_SIZE_GUESS; 
        }
    }
    my $putback = 0;
    while ($size > 0 and $uid = pop @{$self->needsbody}) {
        if (!$self->uids->{$uid}->messageloaded) {
            if (!$self->uids->{$uid}->lock->try) {
                # put back into list, but first... 
                unshift @{$self->needsbody}, 0 unless $putback;
                $putback = 1;
                unshift @{$self->needsbody}, $uid;
                next;
            }
            push @ret, $uid;
            if (defined($self->uids->{$uid}->_size)) {
                $size -= $self->uids->{$uid}->_size;
            } else {
                $size -= 5000; #i think an average message is about 5 k
            }
        }
    }
    return sort { $a <=> $b } (@ret);
}
 
sub load_data {
    my $self = shift;
    # should get this from real server, but lets pretend

    return if ($self->full_path eq '');  # 'root' folder
    async {
        cede;
        $self->status;
    }
}
    
sub imapclient {
    my $self = shift;
    return $self->imappool->get_selected_client($self->full_path);
}    

sub separator {
    my $self = shift;
    return $self->imappool->separator;
}    
 
sub prep_for_destroy {
    my $self = shift;
    if ($self->dec_refcount > 0) {
        return;
    }
    $self->imappool(undef);
    $self->SUPER::prep_for_destroy();
}

sub inc_refcount {
    my $self = shift;
    return $self->refcount (($self->refcount || 0) + 1);
}

sub dec_refcount {
    my $self = shift;
    return $self->refcount (($self->refcount || 0) - 1);
}


sub cache_file_name {
    my $self = shift;
    return 'imap_message_cache.' . sha1_hex($self->full_path);
}

sub close {
    my $self = shift;
    warn "closing";
    if (! $self->selectcount($self->selectcount - 1)) {
        $self->lock->down;
        warn "Building structure to save";
        my %struct = ('uidvalidity' => $self->uidvalidity ,
                      'messages' => {} );  
        my $seq = 0;
        foreach my $message (@{$self->messages}) {
            if ($message->sequence != ++$seq) {
                die "sequence != messages index + 1... major assumption wrong";
            }
            $message->prep_for_destroy();
            $struct{'messages'}{$message->uid} = $message;
        }
        my $filename = $self->cache_file_name(); 
        warn "Storing cache to $filename";
        #print Dumper (\%struct);
        my $status = nstore (\%struct, $filename);
        if (!$status) {
            warn "Failure";
        }
        warn "clearing";
        $self->messages([]);
        $self->uids({});
        $self->needsheaders([]);
        $self->needsbody([]);
        $self->imappool->unselect($self->full_path);
        $self->lock->up;
    }
}

sub select {
    my $self = shift;
    warn "count: " .  $self->full_path . ": " .  ($self->selectcount || 0);
    if ($self->selectcount(($self->selectcount || 0) + 1) == 1) {
        my $client = $self->imapclient;
        
        if (!$client) {
            $self->subscribed(0);
            warn "selecting an unselectable folder";
        } else { 
            my $serverinfo = $client->{'FOLDERS'}{$self->full_path}{'sflags'};
            $self->uidnext($serverinfo->{'UIDNEXT'} || 0);
            
            $self->uidvalidity($serverinfo->{'UIDVALIDITY'});
            
            my $filename = $self->cache_file_name();
            my $cache = {};
            if (-r $filename) {
                warn "Loading cache from $filename";
                $cache = retrieve($filename);
                if (!$cache) {
                    warn "Failure";
                }
                if ($cache->{'uidvalidity'} != $self->uidvalidity) { 
                    warn "uidvalidities don't match, discarding cache";
                    undef $cache;
                }
            }
            my $messagecache = $cache->{'messages'} || {};
            
            $self->recentcount(undef);
            $self->messagecount(undef);
            $self->unseencount(undef);
            # may not be the best way to do this, but we need at least this info to get things started
            my $messages = $client->fetch_with_seq(['1:*'], 'FLAGS UID');
            if (ref $messages eq 'ARRAY') {
                foreach (@$messages) {
                    $self->parse_message($_, $messagecache);
                    cede;
                }
            } else {
                if ($messages != 0) {
                    warn "bad things are happening in " . $self->full_path;
                    warn Dumper ($messages);
                    #warn Dumper ($self);
                }
            }
        }
    }
    return  $self->SUPER::permanentflags(@_);
}

sub recent {
    my $self = shift;
    if (defined($self->recentcount)) {
        return $self->recentcount;
    }
    return $self->SUPER::recent;
}

sub exists {
    my $self = shift;
    if (defined($self->messagecount)) {
        return $self->messagecount;
    }
    return $self->SUPER::exists;
}

sub unseen {
    my $self = shift;
    if (defined($self->unseencount)) {
        return $self->unseencount;
    }
    return $self->SUPER::unseen;
}

sub status {
    my $self = shift;
    $self->lock->down;
    if ($self->is_selectable and ! $self->selectcount) {
        my $status = $self->imappool->status($self->full_path);
        if (!$status) {
            warn "bad status, marking " . $self->full_path . " as unselectable";
            $self->is_selectable(0);
            $self->subscribed(0);
        } else {
            if ($self->full_path ne $status->{'name'}) {
                $self->lock->up;
                die "Bad things, status for " . $status->{'name'} . " but asked for " . $self->full_path;
            }
            $self->uidvalidity($status->{'UIDVALIDITY'});
            $self->uidnext($status->{'UIDNEXT'});
            $self->recentcount($status->{'RECENT'});
            $self->unseencount($status->{'UNSEEN'});
            $self->messagecount($status->{'MESSAGES'});
        }
    }
    $self->lock->up;
    return $self->SUPER::status(@_);
}

sub poll {
    my $self = shift;
    return unless $self->selectcount;
    return unless $self->lock->try;
    async {
        my ($uid) = $self->uids_to_prefetch_headers(undef, 1);
        if ($uid) {
            warn "prefetching $uid";
            $self->uids->{$uid}->load_header();
            $self->uids->{$uid}->lock->up;
        } else {
            ($uid) = $self->uids_to_prefetch_body(undef, 1);
            if ($uid) {
                warn "prefetching body $uid";
                $self->uids->{$uid}->load_message();
                $self->uids->{$uid}->lock->up;
            } else {
                my $client = $self->imapclient;
                if ($client and (time() - $self->lastnoop > $TIME_BETWEEN_NOOPS)) {
                    $client->noop;
                    $self->lastnoop(time());
                }
            }
        }
        my $client = $self->imapclient;
        my $newmessagecount = 0;
        if ($client) {
            my $notifications = $client->notifications;
            foreach (@$notifications) {
                if ($_->{'folder'} ne $self->full_path) {
                    warn "ignoring notification for wrong folder, expected: " . $self->full_path . " recieved: " . $_->{'folder'};
                    print Dumper($_);
                } elsif (defined($_->{'messages'})) {
                    $newmessagecount = $_->{'messages'};
                } elsif ($_->{'destroyed'}) {
                    warn "Expunging message " . $_->{'seq'};
                    my $message = $self->messages->[$_->{'seq'} - 1];
                    if ($message) {
                        if (!$message->has_flag('\Deleted')) {
                            warn "Expunging message without deleted flag";
                            $message->set_flag('\Deleted');
                        }
                        $self->expunge([$_->{'seq'}]);
                        $newmessagecount--;
                    } else {
                        warn "expunging message we don't know about";
                    }
                } elsif ($_->{'flags'}) {
                    my $message = $self->messages->[$_->{'seq'} - 1];
                    if ($message) {
                        $message->store('FLAGS', $_->{'flags'});
                    } else {
                        warn "couldn't set flags for message seq: " .  $_->{'seq'} . " in mailbox " . $self->full_path;
                    }
                } elsif ($_->{'recent'}) {
                    # don't really care about this... presumably we'll see this in flags
                } else {
                    warn "don't know how to handle notification in mailbox " . $self->full_path;
                    print Dumper($_);
                }
            }
            if ($newmessagecount > @{$self->messages}) {
                # may not be the best way to do this, but we need at least this info to get things started
                
                my $messages = $client->fetch_with_seq([ (@{$self->messages} + 1) .':' . $newmessagecount], 'FLAGS UID');
                if (ref $messages eq 'ARRAY') {
                    foreach (@$messages) {
                        $self->parse_message($_);
                        cede;
                    }
                } else {
                    warn "bad things are happening in " . $self->full_path;
                    warn Dumper ($messages);
                    #warn Dumper ($self);
                }
            }
        }
        $self->lock->up;
    }
}

sub get_uids {
    my $self = shift;
    my @return = $self->SUPER::get_uids(@_);
    
    foreach (@return) {
        if (!$_->headerloaded) {
            # will likely be in the list multiple times now...
            # but it will sort itself out down the line
            push @{$self->needsheaders}, $_->uid;
        }
        if (!$_->messageloaded) {
            push @{$self->needsbody}, $_->uid;
        }
    }
    return @return;
}

sub get_messages {
    my $self = shift;
    my @return = $self->SUPER::get_messages(@_);
    
    foreach (@return) {
        if (!$_->headerloaded) {
            # will likely be in the list multiple times now...
            # but it will sort itself out down the line
            push @{$self->needsheaders}, $_->uid;
        }
        if (!$_->messageloaded) {
            push @{$self->needsbody}, $_->uid;
        }
    }
    return @return;
}

sub fetch_messages {
    my $self = shift;
    my ($uid) =  @_;
    
    my @uids_to_fetch = $self->uids_to_prefetch_body($uid, $PREFETCH_BODY_SIZE);
    async {
        my $array = $self->imapclient->fetch(\@uids_to_fetch, 'BODY.PEEK[]') or die
                "couldn't load message uid " . join (', ', @uids_to_fetch) . " from mailbox " . $self->full_path . ": ". $self->imapclient->last_error;
        foreach my $hash (@$array) {
            $self->parse_message($hash);
            cede;
        }
    }
}

sub fetch_headers {
    my $self = shift;
    my ($uid) =  @_;

    my @uids_to_fetch = $self->uids_to_prefetch_headers($uid, $PREFETCH_HEADER_COUNT);
    async {
        my $array = $self->imapclient->fetch(\@uids_to_fetch, 'BODY.PEEK[HEADER] INTERNALDATE RFC822.SIZE ENVELOPE') or die
                "couldn't load headers and stuff uids " . join (', ', @uids_to_fetch)  . " from mailbox " . $self->full_path . ": ". $self->imapclient->last_error;

        foreach my $hash (@$array) {
            $self->parse_message($hash);
            cede;
        }
    }
}


# this might be getting a little too friendly with the ProxyMessage class
# but the function itself needs to be here so we can handle the different
# responses that may be coming back out of order from different fetches

sub parse_message {
    my $self = shift;
    my ($hash, $cache) = @_;
    
    my $uid = $hash->{'UID'} || -1;
    my $message = $self->uids->{$uid};
    if ($hash->{'BODY[]'}) {
        if (!$message) {
            warn "no message foe uid: $uid";
            return;
        }
        my $mime = Email::MIME->new($hash->{'BODY[]'});
        if (!$mime) {
            die "couldn't parse message uid " . $uid . " from mailbox " . $self->mailbox->full_path;
        }
        $message->mime($mime);
        $message->messageloaded(1);
        $message->headerloaded(1);
        $message->header($mime->header_obj);
    } elsif ($hash->{'INTERNALDATE'}) {
        if (!$message) {
            warn "no message for uid: $uid";
            return;
        }
        $message->internaldate($hash->{'INTERNALDATE'});
        if ($message->_size) {
            warn "loaded $uid already";
        }
        $message->_size($hash->{'RFC822.SIZE'});    
        $message->envelope($hash->{'ENVELOPE'});
        $message->header(Email::MIME->new($hash->{'BODY[HEADER]'})->header_obj);
        $message->headerloaded(1);
        if (!$message->header) {
            die "couldn't parse header for message uid " . $uid . " from mailbox " . $self->mailbox->full_path;
        }
    } elsif ($hash->{'seq_id'}) {
        if ($cache) { 
            $message = $cache->{$hash->{'UID'}};
        }
        if ($message) {
            $message->init_flags($hash->{'FLAGS'});
        } else {
            $message = Org::Ruka::ProxyMessage->new(%$hash);
        }
        # lock message so we can unlock it momentarily
        $message->lock->down; 
        
        if ($hash->{'seq_id'} != @{$self->messages} + 1) {
            warn "unexpected sequence id: " . $hash->{'seq_id'} . ' expected: ' . (@{$self->messages} + 1);
        }
        
        $self->add_message($message);
    } else {
        die "weird hash for message uid " . $uid . " from mailbox " . $self->full_path . ' keys: ' . join (' ', keys(%$hash));
    }
    $message->lock->up;
}

#sub add_child {
#    warn "add child not implemented";
#    return 0;
#}

sub create {
    warn "create not implemented";
    return 0;
}

sub reparent {
    warn "reparent not implemented";
    return 0;
}

sub delete {
    warn "delete not implemented";
    return 0;
}

sub append {
    my $self = shift;
    my (@options) = @_;
    
    my $msg = pop @options;
    my ($flags, $date);
    
    ($flags) = grep {ref $_ } @options;
    ($date) = grep { not ref $_} @options;

    #print Dumper (\@_);
    
    my $client = $self->imapclient;
    if ($Org::Ruka::ProxyMessage::cache_write_back) {
        async {
            my $results = $client->append($self->full_path, \$msg, $flags, $date);
        }
        cede;
        return 1;
    } else {
        warn "append: write_through not implemented";
        return 0;
    }
}


# don't actually remove messages, rely on notifications
sub expunge {
    my $self = shift;
    if (Net::IMAP::Server->connection) {
        # comes from client connections, not notifications
        my $only = shift;
        return if $only and not @{$only};  # i don't know why this would happen, but whatever?
        my $client = $self->imapclient;
        
        if ($only and !$client->capability( qr/^UIDPLUS/i)) {
            die "Origin doesn't support UIDPLUS, can't do a partial expunge";
        }
        my @uids;
        if ($only) {
            @uids = ();
            foreach (@$only) {
                my $message = $self->messages->[$_ - 1];
                if ($_ != $message->sequence) {
                    die "sequence != messages index + 1... major assumption wrong";
                }
                push @uids, $_->uid; 
            }
        }
        
        if ($Org::Ruka::ProxyMessage::cache_write_back) {
            async {
                if (!$only) {
                    warn "general expunge";
                    $client->expunge();
                } else {
                    warn "uid expunge: " . join (" ", @uids);
                    $self->uid_expunge(@uids);
                }
            }
        } else {
            die "write_through not implemented";
        }
        return 1;
    } else {
        return $self->SUPER::expunge(@_);
    }
}


1;