package Org::Ruka::ProxyMessage;
use strict;
use warnings;
use Email::MIME;
use Data::Dumper;
use Storable qw(freeze thaw);
use Coro;
use Coro::RWLock;


# if set, writes are asynchronous and assumed to work
# if unset, status will be returned from the origin
# (and copy/append will get UID information)
our $cache_write_back = 1;

use base qw 'Net::IMAP::Server::Message';
Org::Ruka::ProxyMessage->mk_accessors('messageloaded', 'headerloaded', 'envelope', '_size', 'header', 'lock');

sub new {
    my ($class, %args) = @_;
    my $self = bless {}, $class;
    $self->uid($args{'UID'});
    $self->_flags({});
    $self->init_flags($args{'FLAGS'});
    $self->lock(new Coro::Semaphore);
    return $self;
}

sub init_flags {
    my ($self, @flags) = @_;
    if (ref $flags[0] eq 'ARRAY') {
        @flags = @{$flags[0]};
    }
        
    foreach (@flags) {
        $self->_flags->{$_} = 1;
    }
}

sub mime {
    my $self = shift;
    $self->lock->down unless @_;
    if (!@_ and !$self->messageloaded) {
        $self->load_message();
    }
    $self->lock->up unless @_;
    return $self->_mime_accessor(@_);    
}

sub mime_header {
    my $self = shift;
    $self->lock->down;
    if (!$self->headerloaded) {
        $self->load_header();
    }
    $self->lock->up;
    return $self->header;
}


sub internaldate {
    my $self = shift;
    $self->lock->down unless @_;
    if (!@_ and !defined($self->SUPER::internaldate)) {
        $self->load_header();
    }
    $self->lock->up unless @_;
    return $self->SUPER::internaldate(@_);
}

sub mime_select {
    my $self = shift;
    my ($sections, $start, $end, $extras) = @_;
    
    if (!defined($start) and !defined($end) and !defined($extras) and 
        @$sections = 0 and !$self->messageloaded) {
        # looking for size... should be a sub :/
        $self->lock->down;
        if (!defined ($self->_size)) {
            $self->load_header();
        }
        $self->lock->up;
        # is doing a string length on this...
        return " " x $self->_size;
    } else {
        return $self->SUPER::mime_select(@_);
    }
}


sub mime_envelope {
    my $self = shift;
    $self->lock->down;
    if (!defined($self->envelope)) {
        $self->load_header();
    }
    $self->lock->up;
    return $self->envelope;
}

# message lock should be down
sub load_message {
    my $self = shift;
    $self->mailbox->fetch_messages($self->uid);    
    $self->lock->down;
}

# message lock should be down
sub load_header {
    my $self = shift;
    $self->mailbox->fetch_headers($self->uid);
    $self->lock->down;
}

sub prep_for_destroy {
    my $self = shift;
    $self->_flags({});
    $self->sequence(undef);
    $self->SUPER::prep_for_destroy();
}

sub set_flag {
    my $self = shift;
    my ($flag, $silent) = @_;
    
    if (Net::IMAP::Server->connection) {
        # comes from client connections, not notifications
        if ($cache_write_back) {
            async { 
                $self->mailbox->imapclient->add_flags($self->uid, $flag);
            }
        } else {
            die "write_through not implemented";
        }
    }
    return $self->SUPER::set_flag(@_);
}    

sub clear_flag {
    my $self = shift;
    my ($flag, $silent) = @_;
    
    if (Net::IMAP::Server->connection) {
        # comes from client connections, not notifications
        if ($cache_write_back) {
            async { 
                $self->mailbox->imapclient->del_flags($self->uid, $flag);
            }
        } else {
            die "write_through not implemented";
        }
    }
    return $self->SUPER::clear_flag(@_);
}    

# unlike the SUPER, this method does not add a copy to the destination mailbox
# the destination mailbox should be updated by regular server notification on the origin
sub copy {
    my $self = shift;
    my ($mailbox) = @_;
    if ($cache_write_back) {
        async {
            my @results = $self->mailbox->imapclient->copy($self->uid, $mailbox->full_path);
            print Dumper (\@results);
        }
    } else {
        die "write_through not implemented";
    }
    return;
}


1;
